/**
Copyright © 2014 Marko Niinimaki. All Rights Reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
package org.natto;

import java.io.File;
import java.io.FileWriter;
import java.util.UUID;

/**
 * Provides a PGP based security signer/checker for Executor.
 * @author man@cern.ch
 */
public final class ExecSecurity {

    /** instance for singleton. */
    private static ExecSecurity eSecInstance;

    /**
     * This is a singleton. Do not use the constructor.
     */
    private ExecSecurity() {
    }

    /**
     * Use this to get the object.
     * @return ExecSecurity singleton.
     */
    public static synchronized ExecSecurity getInstance() {
        if (eSecInstance == null) {
            eSecInstance = new ExecSecurity();
        }
        return eSecInstance;
    }

    /** what is my security identity? **/
    private static String myIdent = "";
            
    /** is the security on? */
    private boolean securityIsOn = false;

    /**
     * Check if security is on.
     * @return true if it is on
     */
    public boolean securityOn() {
        return securityIsOn;
    }

    /** Turn security off. */
    public void turnOff() {
         securityIsOn = false;
    }

    /** Turn security on.
     * @return diagnostics on failure. "" if successful.
     */
    public String turnOn() {
        String diags = signerOK();
        if ("".equals(diags)) {
            securityIsOn = true;
        }
        return diags;
    }

    /**
     * Return the identity of the signer.
     * @return identity of the signer, "" if unspecified
     */
    public String getMyIdentity() {
        return myIdent;
    }
            
    /** Get signers that we support.
     * @return signers, one on each line
     */
    public String getSigners() {
        String cmdOut = StaticTools.execGet("gpg  --list-keys", false);
        if (cmdOut == null) { cmdOut = ""; }
	String sigNers = "";
        String[] parts = cmdOut.split("\n");
        for (int i=0; i<parts.length; i++) {
	    if (parts[i].startsWith("uid")) {
		String thisLine = parts[i].substring(4);
		thisLine = thisLine.trim();
                sigNers = thisLine + "\n";
	    }
	}
	return sigNers;
    }

    /**
     * Use pgp to add a signature in the file.
     * @param fileName sign this
     */
    public void clearSignFile(String fileName) {
        String cmdOut = StaticTools.execGet("gpg --no-tty --clearsign "
                                            + fileName, true);
        if ("".equals(cmdOut)) {
            File beGone = new File(fileName);
            beGone.delete();
        }
    }

    /**
     * Check the signature, return "" if OK.
     * @param ascFileName file to check
     * @return "" if OK, otherwise error message
     */
    public String verifySignature(String ascFileName) {
        String cmdOut = StaticTools.execGet("gpg --verify "
                                            + ascFileName, true);
        if (cmdOut.indexOf("Good signature") > 0) {
            return "";
        }
        return cmdOut;
    }

    /**
     * Check if the signer program pgp is available and configured in the host.
     * Sets the myIdent variable if OK.
     * @return empty string if OK, otherwise an error message.
     */
    public static String signerOK() {
        String testFileName = UUID.randomUUID().toString() + ".txt";
        File testFile = new File(testFileName);
        File ascFile = new File(testFileName + ".asc");
        try {
            FileWriter fileWr = new FileWriter(testFile);
            fileWr.write(testFileName + "\n");
            fileWr.close();
        } catch (Exception e) { }
        //this command works with most systems that use a modern desktop
        String cmdOut = StaticTools.execGet("gpg --no-tty --clearsign "
                                            + testFileName, true);
        //System.out.println(cmdOut);
        if (cmdOut.indexOf("no terminal at all") > 0) {            
            String x = "";
            String myOs = System.getProperty("os.name");
            if (myOs.startsWith("Mac")) {
                String myDir = System.getProperty("user.dir");
                //some old OSX that does not integrate gpg with desktop
                String commandFileName = testFileName.replace(".txt", ".command");
                File commandFile = new File(commandFileName);
                try {
                   FileWriter fileWr = new FileWriter(commandFile);
                   fileWr.write("cd " + myDir + "\n"
                                + "gpg --clearsign " + testFileName + "\n");              
                   fileWr.close();
                   x = StaticTools.execGet("chmod +x " + commandFileName, true);
                   Thread.sleep(2000);
                   x = StaticTools.execGet("open " + commandFileName, true);
                   Thread.sleep(2000);
                   //wait for the signing to finish
                   boolean goOn = true;
                   while (goOn) {
                       x = StaticTools.execGet("ps", true);
                       if (x.indexOf("clearsign") == -1) {
                           goOn = false;
                       } else {
                           System.out.println("waiting");
                           Thread.sleep(2000);
                       }
                   }
                   commandFile.delete();
                }  catch (Exception e) { }
            } else {
               x = StaticTools.execGet("xterm -e gpg --clearsign "
                                            + testFileName, true);
            }
            System.out.println(x);
	}
        testFile.delete();
        if (ascFile.exists()) {
            //who am I?
            cmdOut = StaticTools.execGet("gpg --verify "
                                          + testFileName + ".asc", true);
            int firstQuotePos = cmdOut.indexOf("\"");
            int lastQuotePos = cmdOut.indexOf("\"", firstQuotePos + 1);
            if (lastQuotePos  > 0) {
                myIdent = cmdOut.substring(firstQuotePos + 1,
                                           lastQuotePos);
            }
            ascFile.delete();
            return "";
        }
        if ("".equals(cmdOut)) {
            return "No response: maybe you do not have the PGP program?";
        }
        if (cmdOut.indexOf("no default") > 0) {
            return "No default pgp key. Please check your PGP configuration.";
        }
        if (!ascFile.exists()) {
            return "PGP could not produce a signed file.";
        }
        return "unspecified error";
    }
}
