/**
Copyright © 2014 Marko Niinimaki. All Rights Reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
package org.natto;
import java.io.*;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;

/**
 * A simple directory-based execution/submission service.
 * Checks periodically if there are .jd files in the directory.
 * If there are, it will execute them if possible.
 * @author man@cern.ch
 */
public class Executor {
    /** My version. **/
    private static String myVer = "0.78";
    /** Directory where we run. **/
    private static String myDirectory = "";

    /** For checking job descriptions. **/
    private enum JdRetType {
       CANNOT_HANDLE, /** No suitable program to execute job.**/
       JD_READ_ERROR, /** Can not read the job description. */
       JD_FORMAT_ERROR, /** Missing fields or not a properties file. */
       JD_OK /** OK. **/
    }

    /** wait this amount of milliseconds between rounds. */
    private static int interval = 5000;
    /** write new timestamp after this many rounds. */
    private static int writeAfter = 6;
    /** info file needs to be younger than this in seconds **/
    private static int maxAge = 300;
    
    /** set to contain the jobs that are being processed in this Executor. */
    private List currentJobs = new ArrayList();
    /** set to contain the jobs that have been submitted by this Executor. */
    private List submittedJobs = new ArrayList();
     /** how many jobs do we let in? set later to be # cores. */
    private int maxRunning = 1;
    /** my id: same as the name of the info file. */
    private static String myId = "";
    /** host info structure. */
    private static HashMap infos = null;
    /** main GUI. We should use only this. SubmitGui is managed by it.
     * Please see swingGUI directory. **/
    private static GenGui onGui = null;
    /** Info file reader/writer. */
    private static org.natto.Infofiles myInfof = null;


    /**A structure of validated stuff from job description. **/
    class ValidatedJd {
        private JdRetType retval; /** Value returned by checker. **/
        private String client = ""; /** submitter. **/
        private String binary = ""; /** binary file. **/
        private String data = ""; /** data file. **/
        private String cmd = ""; /** command line. **/
        private String args = ""; /** extra arguments. **/
        private String target = ""; /** optional target host: run in this. **/

        /** Constructor.
         * @param ret is this OK or has errors
         * @param cli client
         * @param bin binary file name
         * @param dat data file name
         * @param cm command line
         * @param a arguments
         * @param t target host
         **/
        ValidatedJd(JdRetType ret, String cli, String bin, String dat,
                    String cm, String a, String t) {
            retval = ret; client = cli; binary = bin;
            data = dat; cmd = cm; args = a; target = t;
        }
    }

    /**
     * @param args the command line arguments: none, -H filename, -C filename.
     * Creates the Executor object.
     */
    public static void main(String[] args) {
        Executor myExe = null;
        boolean optsOK = false;
        if (args.length == 0) {
            optsOK = true;
            //full gui
            myExe = new Executor(true);
            myExe.start(true);
        }
        if (args.length == 2) {
            String optiOn = args[0];
            String fileName = args[1];
            if (optiOn.equals("-H") || optiOn.equals("-C")) {
                optsOK = true;
                boolean coreCount = false;
                if (optiOn.equals("-C")) { coreCount = true; }
                myExe = new Executor(false);
                String[] suiTables = myExe.getIdsSuitableForTask(fileName,
                                                                 coreCount);
                for (int i=0; i<suiTables.length; i++) {
                    System.out.println(suiTables[i]);
                }
            } else {
                if (optiOn.equals("-S")) { //submit
                    myExe = new Executor(false);
                    myExe.start(false); //we need to create the info file
                    myExe.submitTask(fileName, "", fileName);
                }
            }
        }
        if (!optsOK) {
            System.err.println("Invalid parameters. This program supports \n"
                    + "-H filename list hosts that can execute filename\n"
                    + "-C filename list cores that can execute filename\n"
                    + "-S filename submit task");   
        }
        
    }

    /** give the name of this host's info file.
     * @return info file name
     **/
    public static String getId() {
        return myId;
    }

    /**
     * API-type method: returns hosts/nodes that can execute your task.
     * @param exTension extension like "sh" or a file name like "foo.sh"
     * @param coreWeigth if true, return N*H for H where N is available cores.
     * @return list of info file names
     */
    public String[] getIdsSuitableForTask(String exTension,
                                          boolean coreWeigth) {
        ArrayList okItems = new ArrayList();
        if ((exTension == null) || ("".equals(exTension))) {
            return new String[0];
        }
        if (infos == null) {
            System.err.println("Your infofile data is null. Please check"
                    + " that you have executed checkInfoFiles.");
            return new String[0];
        }
        if (infos.isEmpty()) {
            System.err.println("Your infofile data is empty. Maybe you"
                    + " are not running any executors in this directory.");
            return new String[0];
        }
        int dotPos = exTension.lastIndexOf(".");
        if (dotPos > -1) {
            exTension = exTension.substring(dotPos);
        }
        //iterate infos and search for this extension
        Set<Object> keySet = infos.keySet();
        Iterator keyIt = keySet.iterator();
        while (keyIt.hasNext()) {
             String fName = (String) keyIt.next();
             Properties myProp = (Properties) infos.get(fName);
             //does the property have the extension
             if (myProp.containsKey(exTension)) {
                 if (coreWeigth) {
                     //check available cores
                     String coresStr = myProp.getProperty("cores", "0");
                     int cores = Integer.parseInt(coresStr);
                     String tasksStr = myProp.getProperty("tasks", "0");
                     int tasks = Integer.parseInt(tasksStr);
                     for (int i=0; i<(cores - tasks); i++) {
                         okItems.add(fName);
                     }
                 } else {
                     okItems.add(fName);
                 }
             }
        }
        //return constructed list
        String[] myArray = new String[okItems.size()];
        okItems.toArray(myArray);
        return myArray;
    }
    
    /**
     * Creates a .jd file that can be hopefully executed by some executor.
     * @param myFileName the file to be executed
     * @param myCmd command line
     * @param taskName name for this job
     * @return "" if no errors, otherwise error message
     */
    public String submitTask(String myFileName, String myCmd, String taskName) {
        JobDescription newJobDesc = new JobDescription(myDirectory);
        File file1 = new File(myFileName);
        if (!(file1.canRead())) {
            return "File does not exists or it is not readable.";
        }
        return newJobDesc.writeJobDescription(file1, null, null,
                                              infos, "", myCmd);
    }
            
    /** Constructor.
     * @param useGUI true if you want to run as a real GUI program
     * to use its location
     */
    public Executor(boolean useGUI) {
       new Executor(useGUI, null);
    }
    
    /** Constructor.
     * @param useGUI true if you want to run as a real GUI program
     * @param useThisDir directory for executable, null if you want Executor
     * to use its location. myDirectory class variable is set to this.
     */
    public Executor(boolean useGUI, String useThisDir) {
        String thisdir = useThisDir;
        String myName = "Executor"; //replace by the name of the JAR
        //find out the directory of the jar file that was called.
        if (useThisDir == null) {
            try {
                File jarFile = new File(Executor.class.getProtectionDomain()
                    .getCodeSource().getLocation().toURI().getPath());
                thisdir = jarFile.getParent();
                String tmpName = jarFile.getName();
                if (tmpName.indexOf(".jar") > 0) {
                   myName = myName.replace(".jar", "");
                }
            } catch (Exception e) {
               e.printStackTrace();
            }
        }
        //set the class property
        myDirectory = thisdir;
        //dynamically load the swing gui only if supported
        onGui = new org.natto.TextGui(myName + " " + myVer, thisdir);
        if (useGUI) {
            try {
                Class myGuiClass = Class.forName("org.natto.swingGUI.MainGui");
                //we know it has 2 Strings
                Constructor guiConst = myGuiClass.getConstructor(String.class,
                                                             String.class);
                onGui = (GenGui) guiConst.newInstance(myName + " " + myVer,
                                                              thisdir);
            } catch (Exception e) { e.printStackTrace(); }
        }
        //create infofile reader/writer: it is needed by both GUI and API
        myInfof = new org.natto.Infofiles();
        //using it, create the "info" structure for APIs.
        infos = myInfof.checkInfoFiles(thisdir, "", maxAge);
    }
    
    /**
     * Start the program by checking environment.
     * @param runLoop is true, go to main loop. Otherwise just initialize.
     */
    public void start(boolean runLoop) {    
        onGui.setMessage("Please wait while we look for "
                                             + "command line programs.");
        String userHome = System.getProperty("user.home");
        //we are interested in this file
        String exPropName = userHome + "/.executor.preferences";
        HashMap supportedPrgs =
          org.natto.PrefPrograms.getSupportedPrograms(exPropName);
        System.out.println(supportedPrgs.size()
                + " applications supported by this node.");
        onGui.setMessage("");
        if (runLoop) {
            execLoop(myDirectory, supportedPrgs); //go to main loop
        } else { //just write the info file and re-build infos
            myInfof.writeInfoFile(myDirectory, 0, supportedPrgs);
            infos = myInfof.checkInfoFiles(myDirectory, "", maxAge);
            System.out.println();
        }
    }

    /**
     * Main execution loop.
     * write an info file, observe job descriptions,
     * check number of info files and print it if it has changed,
     * sleep.
     * @param thisdir directory where the program executes.
     * @param prgs programs for info file.
     **/
    public void execLoop(String thisdir, HashMap prgs) {
        int oldNumInfos = 0;
        int roundsSinceWrite = 0;
        boolean goOn = true;
        String iFname = null;
        //get "the" security manager, it is a singleton
        org.natto.ExecSecurity mySecurity =
                        org.natto.ExecSecurity.getInstance();
        while (goOn) {
            boolean securityOn = mySecurity.securityOn();
            onGui.setMessage("");
            if (roundsSinceWrite == writeAfter) { roundsSinceWrite = 0; }
            //put an information file about yourself in the directory
            if (roundsSinceWrite == 0) {
                int numJobsThatIrun = currentJobs.size();
                iFname = myInfof.writeInfoFile(thisdir, numJobsThatIrun, prgs);
            }
            myId = iFname;
            if (iFname == null) {
                onGui.setMessage("Fatal: Cannot write info file.");
                goOn = false;
            }
            roundsSinceWrite++;
            //check info files
            String mySecId = mySecurity.getMyIdentity();
            infos = myInfof.checkInfoFiles(thisdir, mySecId, maxAge);
            Properties myOwnInfo = (Properties) infos.get(iFname);
            maxRunning = Integer.parseInt(myOwnInfo.getProperty("cores", "1"));
            long cores = myInfof.getInfoNumProperty(infos, "cores");
            if (infos.size() != oldNumInfos) {
                onGui.doTell(infos.size() + " hosts and "
                        + cores + " cores available.");
                oldNumInfos = infos.size();
                System.out.println("myinfo " + myOwnInfo);
            }
            //update the gui
            onGui.updateInfosActivate(infos);
            //check if there are running processes
            checkRunningJobs(thisdir);
            //observe job descriptions
            File file = new File(thisdir);
            File[] files = file.listFiles();
            for (int i = 0; i < files.length; i++) {
                String fn = files[i].getName();
                if (fn.endsWith(".run")) {
                    //show info about tasks that the user has submitted
                    TaskExecution.notifyAboutSubmittedJobs(myId, interval,
                                             thisdir, fn, onGui, securityOn);
                }
                if (fn.endsWith(".jd") || (fn.endsWith(".jd.asc"))) {
                    //do we need to check?
                    if (securityOn) {
                        if (fn.endsWith(".jd")) {
                            continue; //we won't run unsecure stuff
                        }
                    }
                    //check to see if this should be run,
                    //and if it is the case, start running it.
                    int extLen = 3;
                    if (fn.endsWith(".jd.asc")) { extLen = 7; }
                    String baseName = fn.substring(0, fn.length() - extLen);
                    String rfName = baseName + ".run";
                    Task onDiskTask = new Task(thisdir + "/" + rfName);
                    if (onDiskTask.getStatus() == Task.IDLE) {
                        //do not let a job in if we are full
                        if (currentJobs.size() < maxRunning) {
                            //security?
                            if (mySecurity.securityOn()) {
                                String verifyErr
                                    = mySecurity.verifySignature(thisdir
                                                                  + "/" + fn);
                                if (!(verifyErr.equals(""))) {
                                    onGui.setMessage(fn + "\ncannot verify");
                                    continue;
                                }
                                onGui.setMessage(fn + " signature ok");
                            }
                            Thread inProcessing = checkStartJdFile(fn, thisdir,
                                                           myId, infos);
                            //System.out.println("Returned");
                            if (inProcessing != null) {
                               org.natto.Task currentJob
                                    = new org.natto.Task(thisdir + "/" + rfName,
                                                     inProcessing, myId);
                               currentJobs.add(currentJob);
                            }  //process was given
                        } //do we have slots?
                    }
                } // .jd
            }
            try {
                Thread.sleep(interval);
            } catch (InterruptedException ex) {
                ex.printStackTrace();
            }
        }
    }

    /** Observe jobs that are running, remove if done.
     * Note: uses currentJobs
     * @param jobDir directory for jobs.
     */
    void checkRunningJobs(String jobDir) {
        for (int i = 0; i < currentJobs.size(); i++) {
           Task currentJob = (Task) currentJobs.get(i);
           //Process currPr = currentJob.jobProcess;
           int jobStat = currentJob.getStatus();
           if ((jobStat == Task.DONE) || (jobStat == Task.NOTIFIED)) {
               currentJobs.remove(i);
           }
       }
    }

    /**
     * Check the jd file and execute it if it needs to be executed by me.
     * @param jdFname job description file name
     * @param thisdir executor work dir
     * @param myId the id of this executor
     * @param infos info file hashes
     * @return process thread if the process was started
     */
     Thread checkStartJdFile(String jdFname, String thisdir,
                               String myId, HashMap infos) {
         //check if job is OK
         Properties myOwnInfo = (Properties) infos.get(myId);
         ValidatedJd checkedJd = canRunJobDescription(jdFname, thisdir,
                                                          myOwnInfo);
         boolean okTotake;
         String jobTarget = checkedJd.target;
         if (!("".equals(jobTarget))) { //is it for someone?
             System.out.println("I am " + myId + " target is " + jobTarget);
             okTotake = jobTarget.equalsIgnoreCase(myId);
         } else {
             okTotake =  org.natto.PoliteScheduler.yesIshouldTakeit(jdFname,
                                                                   myId, infos);
         }
         if (okTotake) { //try to run
             //get the .jd position
             int dotJdpos = jdFname.indexOf(".jd");
             if (dotJdpos < 0) { return null; }
             String baseName = jdFname.substring(0, dotJdpos);
             if (checkedJd.retval == JdRetType.JD_OK) {
                  return uncompressExecuteValidJd(thisdir, checkedJd, baseName);
             }
        }
        return null;
    }


    /** Reads job description and checks if it can be run.
     * @param fname job description file name.
     * @param thisdir directory.
     * @param myProps this host's properties.
     * Note: syntax of the job description file is in
     * https://dl.dropboxusercontent.com/u/919611/executor.html
     * @return a validated job description
     */
    ValidatedJd canRunJobDescription(String fname, String thisdir,
                                     Properties myProps) {
        File jobfile = new File(thisdir, fname);
        if (!((jobfile.exists()) && (jobfile.canRead()))) {
            System.out.println("File " + fname + " not there or readable.");
            return new ValidatedJd(JdRetType.JD_READ_ERROR, "",
                                   "", "", "", "", "");
        }
        //read the file as properties
        Properties jobProp = new Properties();
        try {
            jobProp.load(new FileInputStream(jobfile));
        } catch (Exception e) {
            e.printStackTrace();
            System.out.println("Reading the job description file failed.");
            System.out.println(e.getMessage());
            return new ValidatedJd(JdRetType.JD_READ_ERROR, "",
                                   "", "", "", "", "");
        }
        //check the relevant properties
        String binaryf = jobProp.getProperty("binaryinputfilename");
        if (binaryf == null) {
            System.out.println("binaryinputfilename property missing");
            return new ValidatedJd(JdRetType.JD_FORMAT_ERROR, "",
                                   "", "", "", "", "");
        }
        String cmdLine = jobProp.getProperty("workercommandline");
        if (cmdLine == null) {
            System.out.println("workercommandline property missing");
            return new ValidatedJd(JdRetType.JD_FORMAT_ERROR, "",
                                   "", "", "", "", "");
        }
        //look for the actual command like /usr/bin/python
        //instead of python here. If the file is ELF,
        //the command line is the file.
        if (!(cmdLine.startsWith("INTERNAL"))) {
            cmdLine = getFullCmd(cmdLine, thisdir, binaryf, myProps);
        }
        if (cmdLine.equals("")) {
            System.out.println("Could not run job. Maybe this "
                           + "system has no program to handle file "
                           + binaryf);
            return new ValidatedJd(JdRetType.CANNOT_HANDLE, "",
                                   "", "", "", "", "");
        }
        //System.out.println("cmdline is " + cmdLine);
        String dataf = jobProp.getProperty("datainputfilename");
        //it's ok if its null, data or args are not needed
        String args = jobProp.getProperty("arguments");
        String clientId = jobProp.getProperty("clientid");
        String targetId = jobProp.getProperty("targethost");
        if (targetId == null) { targetId = ""; }
        return new ValidatedJd(JdRetType.JD_OK, clientId,
                               binaryf, dataf, cmdLine, args, targetId);
    }


    /** Using myProps, check the full name of the program that
     * supports the extension of the binary file.
     * @param cmd command
     * @param myDir execution directory
     * @param binFileN binary file name
     * @param myProps properties
     * @return full command string
     * If the file is ELF, the command is the file. But for ELF you
     * have to be Linux. If not, return "".
     */
    static String getFullCmd(String cmd, String myDir, String binFileN,
                             Properties myProps) {
        if (org.natto.StaticTools.fileIsELF(myDir + "/" + binFileN)) {
            String osName = System.getProperty("os.name");
            if (osName.startsWith("Linux")) {
                return cmd;
            } else {
                return "";
            }
        }
        String prgPath = "";
        //get the extension of the binfile
        try {
            int dotPos = binFileN.indexOf(".");
            String binExt = binFileN.substring(dotPos, binFileN.length());
            if (myProps == null) {
                System.out.println("No properties. This should not happen.");
                return "";
            }
            String prgInfo = myProps.getProperty(binExt);
            //since prgInfo can be something like
            //statase C:\Program Files\stata11\statase.exe
            //take everything after first space
            if (prgInfo == null) { return ""; }
            int spacePos = prgInfo.indexOf(" ");
            if (spacePos > -1) {
                prgPath = prgInfo.substring(spacePos + 1,
                                        prgInfo.length());
            }
            //System.out.println(binExt + " " + myId + " " + prgPath);
        } catch (Exception e) { //we just return ""
            e.printStackTrace();
        }
        return prgPath;
    }

    /** Uncompresses data and binary zip files, executes the command line.
     * Data and binary can be ; -separated lists of files or just filename.
     * @param dir the directory where files are.
     * @param myJd a validated job description.
     * @param outBase base (extensionless) name for stdout and err
     * @return the resulting started process thread
     */
    static Thread uncompressExecuteValidJd(String dir,
                                               ValidatedJd myJd,
                                               String outBase) {
        String binfn = myJd.binary;
        String datafn = myJd.data;
        String cmd = myJd.cmd;
        String args = myJd.args;
        //check if the binary files exist and if they need to be unzipped.
        String[] filesinZips = new String[0];
        if (binfn.toLowerCase().endsWith(".zip")) {
            filesinZips = checkUnzipFileStr(dir, binfn);
            //check here if there is a supported file in the zip
        }
        if (filesinZips == null) { return null; } //we cannot continue
        if (datafn != null) {
            if (datafn.toLowerCase().endsWith(".zip")) {
                filesinZips = checkUnzipFileStr(dir, datafn);
            }
        }

        return TaskExecution.execCommand(dir, cmd, binfn, args,
                                         outBase, interval, myId);
    }


    /** Auxiliary to prevent code mess.
     * @param dir directory
     * @param semicSep semicolon separated string
     * @return array of the filenames in zip
     */
    private static String[] checkUnzipFileStr(String dir, String semicSep) {
        ArrayList combinedZipFileList = new ArrayList();
        if ((semicSep == null) || ("".equals(semicSep.trim()))) {
            return new String[0];
        }
        String[] myFilenames = semicSep.split(";");
        for (int i = 0; i < myFilenames.length; i++) {
            File myFile = new File(dir + "/" + myFilenames[i]);
            if (!(myFile.exists())) {
                System.out.println("File " + myFilenames[i] + " not in " + dir);
                return null;
            }
            String[] filesInZip = org.natto.StaticTools.unZip(dir + "/"
                    + myFilenames[i]);
            combinedZipFileList.addAll(Arrays.asList(filesInZip));
        }
        return (String[]) combinedZipFileList.toArray(new String[0]);
    }

}
