/**
Copyright © 2014 Marko Niinimaki. All Rights Reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
package org.natto;

import java.io.*;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.UUID;

/**
 * Write/read info files for the simple execution service.
 * Note: this class could be a singleton because you
 * will only need one instance of it.
 * Note: uses and creates an instance of the Sysinfo class.
 * @author man@cern.ch
 */
public class Infofiles {
    /** Used so that the info file generator does not generate a new
     * one every time it is called. */
    static final String internalUnique = UUID.randomUUID().toString();
    /**make an sysinfo object, since its constructor gets CPU etc.*/
    private org.natto.Sysinfo mySysinfo = new org.natto.Sysinfo();

     /** Write info about this host to a file with a unique name.
     * @param dirName name of directory where the file will be written
     * @param jobsInMe number of jobs that I run
     * @param prgs programs that are listed as software environments
     * @return infoFileName if successful, null if not
    **/
    public String writeInfoFile(String dirName, int jobsInMe, HashMap prgs) {
        String infoFname = getUniqFileName();
        //System.out.println("infoFname " + infoFname);
        //get some basic info from the Java environment
        String os =  System.getProperty("os.name");
        String osVer =  System.getProperty("os.version");
        //and the information system
        int cores = mySysinfo.getCores();
        int mhz = mySysinfo.getMhz();
        int memMeg = mySysinfo.getMemMeg();
        long freeDiskMegs = mySysinfo.getFreeDiskMeg();
        String hostName = mySysinfo.getHostName();
        String userName = System.getProperty("user.name");
        ExecSecurity mySecurity = ExecSecurity.getInstance();
        String acceptedSigners = "*";
        if (mySecurity.securityOn()) {
            acceptedSigners = mySecurity.getSigners();
            acceptedSigners = acceptedSigners.trim();
            acceptedSigners = acceptedSigners.replace("\n", " -- ");
        }
        //get the date too so that we indicate change in the file
        //windows: delete the old info before you continue
        try {
            File myFile = new File(dirName + "/" + infoFname);
            if (myFile.exists()) {
                myFile.delete();
            }
        } catch (Exception e) {
             System.out.println("Cannot delete old info file");
             e.printStackTrace();
        }
        try {
           PrintWriter writer = new PrintWriter(dirName + "/" + infoFname,
                                                 "UTF-8");
           long epoch = StaticTools.getEpoch();
	   writer.println("timestamp=" + epoch);
           writer.println("tasks=" + jobsInMe);
	   writer.println("os=" + os);
	   writer.println("os_version=" + osVer);
	   writer.println("cores=" + cores);
	   writer.println("mhz=" + mhz);
	   writer.println("mmem=" + memMeg);
	   writer.println("diskfree=" + freeDiskMegs);
	   writer.println("name=" + hostName);
           writer.println("user=" + userName);
           writer.println("signers=" + acceptedSigners);
	   Set<Object> keySet = prgs.keySet();
           Iterator keyIt = keySet.iterator();
           while (keyIt.hasNext()) {
               String key = (String) keyIt.next();
	       String[] prgInfo = (String[]) prgs.get(key);
	       String prgName = prgInfo[0];
	       String prgCmd = prgInfo[1];
	       writer.println(key + "=" + prgName + " " + prgCmd);
	   }
	   writer.close();
	} catch (Exception e) { e.printStackTrace(); }
        return infoFname;
    }

    /** Get a unique name for info file. Try hardware info.
     * If no luck, use a random id.
     * @return unique id that can be used as an info file name
     */
    private static String getUniqFileName() {
       String userName =
               System.getProperty("user.name").trim().replace(" ", "");
       String uniqueID = internalUnique + ".info";
       try {
	    //super simple: get eth0 on Unix if there is one.
            File ethAddf = new File("/sys/class/net/eth0/address");
            if (ethAddf.exists()) {
		BufferedReader myReader
		    = new BufferedReader(
			    new FileReader(ethAddf));
		String ethMac = myReader.readLine();
		myReader.close();
		ethMac = ethMac.replace(":", "");
		return ethMac + userName + ".info";
	    }
       } catch (Exception e) { e.printStackTrace(); }

       //previous did not work? try windows registry
       String cmd = "reg query HKLM\\SYSTEM\\CurrentControlSet\\Control\\Class\\{4D36E972-E325-11CE-BFC1-08002BE10318} /s";
       String output = org.natto.StaticTools.execGet(cmd, false);
       //parse the output to obtain NetCfgInstanceId
       String[] lines = output.split("\\n");
       for (int i = 0; i < lines.length; i++) {
            String line = lines[i];
            if (line.contains("NetCfgInstanceId")) {
                //it will look like:
                //NetCfgInstanceId REG_SZ {DCB14C61-690D-46F7-8A89-150432FA5C44}
                //take the stuff after last -
                int dashPos = line.lastIndexOf('-');
                if (dashPos > 0) {
                    String myUni = line.substring(dashPos + 1);
                    myUni = myUni.trim();
                    myUni = myUni.substring(0, myUni.length() - 1);
                    return myUni + userName + ".info";
                }

            }
       }
       //next: in OSX, run ifconfig | grep ether | cut -d' ' -f 2 | head -1
       cmd = "ifconfig";
       output = org.natto.StaticTools.execGet(cmd, false);
       lines = output.split("\\n");
       for (int i = 0; i < lines.length; i++) {
            String line = lines[i];
            int etherPos = line.indexOf("ether");
            if ((etherPos > -1) && (line.length() > etherPos + 6)) {
                String physAdd = line.substring(etherPos + 6);
                physAdd = physAdd.replace(":", "");
                physAdd = physAdd.trim();
	        System.out.println(physAdd);
                return physAdd + userName + ".info";
            }
       }

       //nothing worked, return unique
       return uniqueID;
   }

    /**
     * Checks the info files in the directory.
     * @param dirName in this directory
     * @param secIdent security common name. Used to determine what hosts
     * accept this user. "" if anything goes.
     * @param seconds only return those newer that this
     * @return a hash map where keys are filenames and values properties
     * of extensions: dC9ojpbGman.info -> timestamp 10000 etc
     */
    public static HashMap checkInfoFiles(String dirName, String secIdent,
                                         int seconds) {
        HashMap hm = new HashMap();
        File folder = new File(dirName);
	File[] listOfFiles = folder.listFiles();
	if (listOfFiles == null) {
            return hm;
        }
        for (int i = 0; i < listOfFiles.length; i++) {
            if (listOfFiles[i].isFile()) {
		String fName = listOfFiles[i].getName();
                //we cannot rely on file dates since
                //the underlying file system can be
                //syncing. Read timestamp from the file.
                if (fName.endsWith(".info")) {
                    try {
                        boolean secOK = true;
		        Properties myProp = new Properties();
                        myProp.load(new FileInputStream(listOfFiles[i]));
                        String acceptedSigners = myProp.getProperty("signers");
                        if ("".equals(secIdent) && 
                                !("*".equals(acceptedSigners))) {
                            //System.out.println(fName + " requires signatures "
                            //        + "but we do not sign.");
                            secOK = false;                           
                        }
                        //if we submitter signs, check if the signature is OK
                        if (secIdent.length() > 0) {
                            if ("*".equals(acceptedSigners)) {
                                secOK = true; //this host accepts anything
                            } else {
                                int myPosInSec = 
                                        acceptedSigners.indexOf(secIdent);
                                secOK = (myPosInSec > -1); //I am accepted
                            }
                        }
                        String tStamp = myProp.getProperty("timestamp");
                        long timeFile = Long.parseLong(tStamp);
                        //compare w current time

                        long secondsSinceEpoch =
                            StaticTools.getEpoch();
                        if (secondsSinceEpoch - timeFile < seconds) {
			    //file is fresh
                            //System.out.println(fName + " ok");
                            if (secOK) {
                                hm.put(fName, myProp);
                            } else {
                                System.out.println(fName + " rejected by "
                                                         + "security policy.");
                            }
			} else {
                            //System.out.println(fName + " too old");
			}
		    } catch (Exception e) { e.printStackTrace(); }
                }
            }
        }
        return hm;
    }

    /**
     * Get a property from file.
     * @param hm hash map of info files
     * @param infoFileName name of file
     * @param propName name of property
     * @return property value or null if file or property does not exist
     */
    public static String getHostProperty(HashMap hm, String infoFileName,
                                  String propName) {
        if (!hm.containsKey(infoFileName)) { return null; }
        Properties myProp = (Properties) hm.get(infoFileName);
        if (!myProp.containsKey(propName)) { return null; }
        return myProp.getProperty(propName);
    }

    /**
     * Search for a sum property in the valid info files presented by the hash.
     * @param myHm hash map of the info files
     * @param mySearch search for this
     * @return cumulative sum of the property values
     */
    public long getInfoNumProperty(HashMap myHm, String mySearch) {
        Set<Object> keySet = myHm.keySet();
        Iterator keyIt = keySet.iterator();
        long cumSum = 0;
        while (keyIt.hasNext()) {
            try {
                String fName = (String) keyIt.next();
                Properties myProp = (Properties) myHm.get(fName);
                //get the thing the user wants
                cumSum = cumSum
                        + Long.parseLong(myProp.getProperty(mySearch, "0"));
            } catch (Exception e) { }
        }
        return cumSum;
    }

    /** Get supported extensions as a string from info files.
     * @param myHm a hash map of names and properties
     * @return a string of space separated extensions
    */
    public static String getSupportedExt(HashMap myHm) {
	String ret = "";
        HashMap uniQr = new HashMap(); //we do not list .py many
        //times though it can appear in many files
	//just get properties from each key and from each property
        //stuff that start with a period
   	Set<Object> keySet = myHm.keySet();
        Iterator keyIt = keySet.iterator();
        while (keyIt.hasNext()) {
            try {
                String fName = (String) keyIt.next();
		//System.out.println("XXX " + fName);
                //get the properties
                Properties myProp = (Properties) myHm.get(fName);
                //get all keys
		Enumeration myEnum = myProp.propertyNames();
                while (myEnum.hasMoreElements()) {
		    String pKey = (String) myEnum.nextElement();
                    if (pKey.startsWith(".")) {
			uniQr.put(pKey, pKey);
                    }
		}
	    } catch (Exception e) { e.printStackTrace(); }
	}
        //get keys from uniquer and put them in a string
	Set<Object> ukeySet = uniQr.keySet();
        keyIt = ukeySet.iterator();
        while (keyIt.hasNext()) {
            String k = (String) keyIt.next();
	    String uKey = (String) uniQr.get(k);
            ret = ret + " " + uKey;
	}
        ret = ret.trim();
        return ret;
    }

    /** Get the program name for the given extension from info files.
     * Example: .py -> /usr/bin/python
     * Please note that Windows prg calls may look like:
     * "C:\\Program Files\\My Application\\thisandthat.exe" /T
     * @param ext the extension.
     * @param myHm a hash map of names and properties
     * @return program name (full path)
    */
    public static String getPrgForExt(String ext, HashMap myHm) {
        String extPrg = ""; //find and return this
        //like python /usr/bin/python
        Set<Object> keySet = myHm.keySet();
        Iterator keyIt = keySet.iterator();
        while (keyIt.hasNext()) {
            try {
                String fName = (String) keyIt.next();
                //get the properties
                Properties myProp = (Properties) myHm.get(fName);
                String val = myProp.getProperty(ext);
                if (val != null) {
		    //take the first part
                    String[] valParts = val.split(" ");
                    extPrg = valParts[0];
                    break;
		}
	    } catch (Exception e) { e.printStackTrace(); }
	}
        return extPrg;
    }
}
