/**
Copyright © 2014 Marko Niinimaki. All Rights Reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
package org.natto;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * Represents a job description file.
 * @author man@cern.ch
 */
public class JobDescription {
    private String workDir = "";

    /**
     * Constructor.
     * @param sharedDir working directory
     */
    public JobDescription(String sharedDir) {
        workDir = sharedDir;
    }

    /**
     * Write a job description file, return diagnostics if fails.
     * @param file1 file name of the main file, mandatory
     * @param file2 file name of the optional file
     * @param zipFile file name of the optional zip file, alternate to file2
     * @param prgHash program hash
     * @param clineExtra extra command line arguments
     * @param toHost host name, optional
     * @return submit message, contains "submitted" if OK
     */
    public String writeJobDescription(File file1, File file2, File zipFile,
            HashMap prgHash,
            String clineExtra,
            String toHost) {
	//unique name for files and control files
	String myUniq = java.util.UUID.randomUUID().toString();
        String jobFileName = workDir + "/" + myUniq + ".jd";
        boolean submitOk = true;
        try {
	    PrintWriter writer = new PrintWriter(jobFileName,
                                                 "UTF-8");
	    //get names
	    String myId = org.natto.Executor.getId();
	    writer.println("clientid=" + myId);
	    String f1Name = file1.getName();
	    writer.println("binaryinputfilename=" + f1Name);
            writer.println("jobname=" + f1Name);
	    //get the command that corresponds with the file's extension
	    String myExt = org.natto.StaticTools.getFileExt(f1Name);
            String wcommandLine;
            if (myExt == null) { //internals and Unix programs: ext can be null
                if (f1Name.startsWith("INTERNAL")) {
                    wcommandLine = "workercommandline=" + f1Name;
                } else {
		    wcommandLine = "workercommandline=./" + f1Name;
                }
            } else {
	        String myPrg = org.natto.Infofiles.getPrgForExt(myExt, prgHash);
		wcommandLine = "workercommandline=" + myPrg + " " + f1Name;
	    }
            writer.println(wcommandLine);
            if (!("".equals(clineExtra))) {
		writer.println("arguments=" + clineExtra);
	    }

            ArrayList dataInputAl = new ArrayList();
            if (zipFile != null)  { dataInputAl.add(zipFile.getName()); }
            if (file2 != null) { dataInputAl.add(file2.getName()); }
            String diStr = org.natto.StaticTools.join(dataInputAl, ";");
            if ((diStr != null) && (diStr.length() > 0)) {
		writer.println("datainputfilename=" + diStr);
	    }
            if (toHost != null) {
                if (!("".equals(toHost))) {
                    writer.println("targethost=" + toHost);
                }
            }
	    writer.close();
	} catch (Exception e) { submitOk = false; }
        //sign the file if security is on
        org.natto.ExecSecurity mySecurity
                = org.natto.ExecSecurity.getInstance();
        if (mySecurity.securityOn()) {
            mySecurity.clearSignFile(jobFileName);
        }
	if (submitOk) {
	   return "Task " + myUniq + " submitted.";
        } else {
	    return "Sorry, submit failed.";
	}
    }

    /**
     * Create a job description for an internal command like XMLLS.
     * @param command internal command
     * @param param parameter
     * @param toHost this must be executed in this host
     * @return diagnostics
     */
    public String writeInternalCommand(String command, String param,
                                       String toHost) {
        String retMe = writeJobDescription(
                            new File("INTERNAL:" + command), null,
                            null, null, param,
                            toHost);
        return retMe;
    }

}
