/**
Copyright © 2014 Marko Niinimaki. All Rights Reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
package org.natto;
import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;

/**
 * Finding/reading/writing a preferences file.
 * @author man@cern.ch
 */
public class PrefPrograms {

    /** Used as external variable in traverseSeek. */
    private static HashMap traverseHm = new HashMap();

    /**
     * Test program.
     * @param args the command line arguments
     */
    public static void nomain(String[] args) {
        String userHome = System.getProperty("user.home");
        //we are interested in this file
        String exPropName = userHome + "/.executor.preferences";
        HashMap h = getSupportedPrograms(exPropName);
        System.out.println(h.size() + " supported.");
    }


    /**
     * The only thing you will want to call from here.
     * @param exPropName property file name.
     * @return HashMap of ext -> [program,fullpath] like
     * .py -> [python,/usr/bin/python]
     */
    public static HashMap getSupportedPrograms(String exPropName) {
        HashMap hm = new HashMap();
        // Read strings of type extensions_x from preferences file.
        // Make two arrays.
        ArrayList<String> myPrgList = new ArrayList(); //programs
        ArrayList<String> myExtList = new ArrayList(); //exts
        //init the pref file
        doPreferencesFile(exPropName);
        //now that it should exist, read it
        try {
            Properties exProp = new Properties();
            exProp.load(new FileInputStream(exPropName));
	    //iterate over values to find programs the user is
	    //interested in
	    Set<Object> keySet = exProp.keySet();
            Iterator keyIt = keySet.iterator();
            while (keyIt.hasNext()) {
		String key = (String) keyIt.next();
                String vals = (String) exProp.get(key);
                String[] singleVals = vals.split(",\\s*");
		for (int i = 0; i < singleVals.length; i++) {
                    String val = singleVals[i];
                    if (key.startsWith(".")) {
                        myPrgList.add(val);
                        myExtList.add(key);
                    }
		}
	    }
	} catch (Exception e) { e.printStackTrace(); }
        //call the sub that builds the supported program info
        //for writing it into an info file.
        hm = checkSupportedPrograms(myPrgList, myExtList);
        return hm;
    }

    static HashMap checkSupportedPrograms(List wantedPrograms, List exts) {
	String osName = System.getProperty("os.name");
        HashMap retMe = null;
	if (osName.startsWith("Windows")) {
            retMe = checkSupportedProgramsWindows(wantedPrograms, exts);
	} else {
	    retMe = checkSupportedProgramsUnix(wantedPrograms, exts);
	}
        return retMe;
    }

    static HashMap checkSupportedProgramsUnix(List wantedPrograms, List exts) {
        return checkByDirectCall(wantedPrograms, exts, "which");
    }

    static HashMap checkByDirectCall(List wantedPrograms, List exts,
                                     String callMethod) {
        //the structure of this: extension -> prgname, howToCall
        HashMap hm = new HashMap();
        //check which ones of these are executable by the system
	for (int i = 0; i < wantedPrograms.size(); i++) {
	    String ext = (String) exts.get(i);
	    String prg = (String) wantedPrograms.get(i);
            //this can be a program and argument like java -jar, so split
            String[] prgArgs = prg.split(" ", 2);
            prg = prgArgs[0];
            String args = "";
            if (prgArgs.length > 1) {
                args = " " + prgArgs[1].trim();
            }

            String cmd =  prg + " -version";
            if (callMethod.equals("which")) {
                //use which to get a full path to the prg
                cmd =  "which " + prg;
            }

            String osname = System.getProperty("os.name");
            //we cannot let the system try cmd because it
            //does not return
            if (osname.startsWith("Windows")
                    && prg.toLowerCase().equals("cmd")) {
                 hm = addInHash(hm, ext, prg, "cmd" + args);
            } else {
	       try {
                   Process process = Runtime.getRuntime().exec(cmd);
                   process.waitFor();
                   BufferedReader stdInput =
                        new BufferedReader(
                        new InputStreamReader(process.getInputStream()));
                   String s = stdInput.readLine();
                   if (s != null) {
                        //if we use the -version, we do not get the full path,
                        //just use the prg name
                        if (callMethod.equals("-version")) {
                            s = prg;
                        }
                        //if we used "which", the reponse can be
                        //"no x in p1 p2.." where p's are path elements
                        if (!(s.startsWith("no "))) {
                            hm = addInHash(hm, ext, prg, s + args);
			}
                    }
                } catch (Exception e) {
                    //e.printStackTrace();
                }
            }
        }
        return hm;
    }

    static HashMap checkSupportedProgramsWindows(List wantedPrograms,
                                                 List exts) {
        //check which ones of these are just executables by the system
        traverseHm = checkByDirectCall(wantedPrograms, exts, "-version");
        //if this is Windows, check programs in Program Files
        //first find where PF is by
        String query =
            "reg query HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion";
        String prgFiles = "C:\\Program Files";
        try {
            Process process = Runtime.getRuntime().exec(query);
            BufferedReader stdInput = new BufferedReader(new
                 InputStreamReader(process.getInputStream()));
            String s = "";
            while ((s = stdInput.readLine()) != null) {
                if (s.indexOf("ProgramFilesDir") > 0) {
                    //take anything after REG_SZ and trim
                    int regszPos = s.indexOf("REG_SZ");
                    if (regszPos > 0) {
                        String after = s.substring(regszPos + 6,
                                               s.length());
                        after = after.trim();
                        prgFiles = after;
                    }
		}
            } //read input
            //traverse the directory tree of
            //program files and look for .exe files
            traverseSeek(new File(prgFiles), wantedPrograms, exts);
        } catch (Exception e) { e.printStackTrace(); }
        return traverseHm;
    }


    /**search in-depth for programs, update traverseHm. **/
    static void traverseSeek(File node, List programs, List exts) {
        String nodeName = node.getAbsoluteFile().getName();
	nodeName = nodeName.toLowerCase(); //windows
        //System.out.println(nodeName);
        if (node.isDirectory()) {
            String[] subNode = node.list();
            if (subNode != null) {
                for (int i = 0; i < subNode.length; i++) {
                    String fileName = subNode[i];
                    traverseSeek(new File(node, fileName), programs, exts);
                }
            }
        } else {
	    //check for / program .exe for any prg here
            for (int i = 0; i < programs.size(); i++) {
                //windows is case-insensitive
		String prgName = (String) programs.get(i);
                String[] prgArgs = prgName.split(" ", 2);
                prgName = prgArgs[0];
                String args = "";
                if (prgArgs.length > 1) {
                    args = " " + prgArgs[1].trim();
                }
		prgName = prgName.toLowerCase();
		if (nodeName.indexOf(prgName + ".exe") == 0) {
                    String thisExt = (String) exts.get(i);
                    String path = node.getAbsolutePath();
                    //damn windows brainless ****. If this is written
                    //in a properties file without double-encoding backslashes,
                    //it cannot be read from there.
                    path = path.replace("\\", "\\\\");
                    path = "\"" + path + "\"";
		    System.out.println("found " + path);
                    path = path + args;
		    traverseHm = addInHash(traverseHm, thisExt, prgName, path);
		}
	    }
	}
    }

    //add in a key -> String[2] structure
    static HashMap addInHash(HashMap orig, String ext, String prg,
                             String full) {
        //System.out.println("Adding " + prg);
        String[] proCom = new String[2];
        proCom[0] = prg;
        proCom[1] = full;
        orig.put(ext, proCom);
        return orig;
    }

    /** Read a preferences file if it exists. If not, create and write.
     * @param exPropName file name
     * @return the preferences
     **/
    static Properties doPreferencesFile(String exPropName) {
        Properties exProp = new Properties();
        //fill in some default values
        exProp.setProperty(".py", "python");
        exProp.setProperty(".pl", "perl");
        exProp.setProperty(".r", "R --vanilla <");
        exProp.setProperty(".do", "stataSE /e, stata /e");
        exProp.setProperty(".tex", "pdflatex -halt-on-error");
        exProp.setProperty(".sh", "bash");
        exProp.setProperty(".mpi", "mpirun");
        exProp.setProperty(".jar", "java -jar");
        exProp.setProperty(".bat", "cmd /q /c");
        exProp.setProperty(".cmd", "cmd /q /c");
        exProp.setProperty(".exe", "cmd /q /c");
	boolean readOK = true;
        //try to read settings from the file
        try {
            File exPropF = new File(exPropName);
            if (exPropF.exists()) {
		exProp.load(new FileInputStream(exPropName));
	    }
	} catch (Exception e) {
            //e.printStackTrace();
	}
        //write them back
	try {
	    FileOutputStream exPropFOS =
		new FileOutputStream(exPropName);
	    exProp.store(exPropFOS, "Please exit the executor "
                    + "program before making changes in this file!");
	    exPropFOS.flush();
	    exPropFOS.close();
	} catch (Exception e) {
	    e.printStackTrace();
	}
	return exProp;
    }

}
