/**
Copyright © 2014 Marko Niinimaki. All Rights Reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
package org.natto;
import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collection;
import java.util.Iterator;
import java.util.Properties;
import java.util.TimeZone;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

/** Methods that are called without making an object.*/
public class StaticTools {

    /** Copy file to thisDir unless the file is there already.
     * @param file the file to copy
     * @param thisDir into this directory
     * @return true if success
    */
    public static boolean copyFileIfNeeded(File file, String thisDir) {
        boolean copyOK = true;
        //check file names
        try {
            String fullFn = file.getCanonicalPath();
            String fnOnly = file.getName();
            File futureFile = new File(thisDir, fnOnly);
            String fullfFn = futureFile.getCanonicalPath();
            if (!(fullfFn.equals(fullFn))) {
                copyFileUsingStream(file, futureFile);
	    }
	} catch (Exception e) { copyOK = false;
                                e.printStackTrace(); 
        }
	return copyOK;
    }



    static void copyFileUsingStream(File source, File dest)
                                        throws IOException {
        InputStream is = null;
        OutputStream os = null;
        try {
            is = new FileInputStream(source);
            os = new FileOutputStream(dest);
            byte[] buffer = new byte[1024];
            int length;
            while ((length = is.read(buffer)) > 0) {
                os.write(buffer, 0, length);
            }
        } finally {
            if (is != null) is.close();
            if (os != null) os.close();
        }
    }

    /**
     * Simple sub to join things as a string.
     * @param s collection
     * @param delimiter separated by
     * @return joined string
     */
    public static String join(Collection s, String delimiter) {
        StringBuffer buffer = new StringBuffer();
        Iterator iter = s.iterator();
        while (iter.hasNext()) {
            buffer.append(iter.next());
            if (iter.hasNext()) {
                buffer.append(delimiter);
            }
        }
        return buffer.toString();
    }

    public static long getEpoch() {
        Calendar calendar = Calendar.getInstance(
                                TimeZone.getTimeZone("UTC"));
        long secondsSinceEpoch = calendar.getTimeInMillis() / 1000L;
	return secondsSinceEpoch;
    }

     /** Simple unzipper.
     * @param zipFile file name.
     * @return list of file names.
    */
    public static String[] unZip(String zipFile) {
        return unZip(zipFile, false);
    }

    /** Simple unzipper.
     * @param zipFile file name.
     * @param listOnly do not unzip, just give a list
     * @return list of file names.
    */
    public static String[] unZip(String zipFile, boolean listOnly) {
	ArrayList zipNames = new ArrayList(); //a list of file names in the Zip
        byte[] buffer = new byte[1024];
        try {
             System.out.println("Checking " + zipFile);
    	     //get the zip file content
    	     ZipInputStream zis =
    		 new ZipInputStream(new FileInputStream(zipFile));
    	     //get the zipped file list entry
    	     ZipEntry ze = zis.getNextEntry();
    	     while (ze != null) {
    	          String fileName = ze.getName();
		  zipNames.add(fileName);
                  if (!listOnly) {
                    File newFile = new File(fileName);
                    //create all non existant folders
                    //else: FileNotFoundException for compressed folder
                    //new File(newFile.getParent()).mkdirs();
                    FileOutputStream fos = new FileOutputStream(newFile);
                    int len;
                    while ((len = zis.read(buffer)) > 0) {
       		      fos.write(buffer, 0, len);
                    }
                    fos.close();
                  }
                  ze = zis.getNextEntry();
	     }
	     zis.closeEntry();
	     zis.close();
	} catch (IOException ex) {
	    ex.printStackTrace();
	}
	String[] fileArr = new String[zipNames.size()];
	zipNames.toArray(fileArr);
	return fileArr;
    }

    /** Checks if this is an ELF executable.
     * @param file file to check
     * @return true if the file is ELF
    */
    public static boolean fileIsELF(File file) {
	if (!(file.exists())) return false;
        byte[] elfBytes = { 0x7F, 0x45, 0x4C, 0x46 };
	byte[] fileBytes = new byte[4];
	try {
	    //read the 4 first bytes
	    FileInputStream fis = new FileInputStream(file);
            fis.read(fileBytes, 0, 4);
	} catch (Exception e) { e.printStackTrace(); return false; }
        //the bytes should read .ELF
        return Arrays.equals(elfBytes, fileBytes);
    }

    /**
     * Check if file is ELF.
     * @param fileName name of the file.
     * @return true if file is ELF.
     */
    public static boolean fileIsELF(String fileName) {
	File file = new File(fileName);
        return fileIsELF(file);
    }

     /** Execute a command and get its result.
      * @param exeCmd the command to execute
      *
      * @return command output
     */
    static String execGet(String exeCmd, boolean stdErrToo) {
	String outPut = "";
	try {
	    Process process = Runtime.getRuntime().exec(exeCmd);
	    StreamReader reader
                     = new StreamReader(process.getInputStream());
            StreamReader errReader
                     = new StreamReader(process.getErrorStream());
	    reader.start();
            errReader.start();
	    process.waitFor();
	    reader.join();
            errReader.join();
	    outPut = reader.getResult();
            if (stdErrToo) {
                outPut = outPut + errReader.getResult();
            }
	} catch (Exception e) { e.printStackTrace(); }
	return outPut;
    }

    static class StreamReader extends Thread {
        private InputStream is;
        private StringWriter sw = new StringWriter();

        public StreamReader(InputStream is) {
            this.is = is;
        }

        public void run() {
            try {
                int c;
                while ((c = is.read()) != -1) { sw.write(c); }
            } catch (IOException e) { }
        }

        public String getResult() {
            return sw.toString();
        }
    }

    /**
     * Get properties from a file.
     * @param fileName name of file
     * @return properties in the file
     */
    public static Properties getPropertiesFromFile(String fileName) {
        Properties myProp = new Properties();
        File propFile = new File(fileName);
        try {
            FileInputStream fis = new FileInputStream(propFile);
            myProp.load(fis);
            fis.close();
        } catch (Exception e) {}
        return myProp;
    }

    /**
     * Get a file's extension.
     * @param fname file name
     * @return extension
     */
    public static String getFileExt(String fname) {
        String myExt = null;
        int dotPos = fname.indexOf(".");
	if (dotPos > -1) {
	    myExt = fname.substring(dotPos, fname.length());
	}
        return myExt;
    }

    /**
     * Give the contents of a text file.
     * @param fileName name of file
     * @param maxLines if >0, read max this number of lines
     * @return null if not successful, otherwise file contents
     */
    public static String getFileContent(String fileName, int maxLines) {
        File readFile = new File(fileName);
        if (maxLines == 0) { maxLines = Integer.MAX_VALUE; }
        String retMe = "";
        int curLine = 0;
        if (readFile.exists() && readFile.canRead()) {
            try {
                FileInputStream fis = new FileInputStream(readFile);
                BufferedReader br = new BufferedReader(
                        new InputStreamReader(fis));
                String myLine = null;
                while ((myLine = br.readLine()) != null) {
                     curLine++;
                     retMe = retMe + myLine;
                     if (curLine >= maxLines) { break; }
                }
                br.close();
                fis.close();
            } catch (Exception e) {
                e.printStackTrace();
                return null;
            }
        } else {
            return null;
        }
        return retMe;
    }

    public static String getFileContent(String fileName) {
        return getFileContent(fileName, 0);
    }
}
