/**
Copyright © 2014 Marko Niinimaki. All Rights Reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
package org.natto;

import java.io.*;

/**
 * Sysinfo elements for the simple execution service.
 * Find the number of cores, CPU frequency etc.
 * @author marko
 */
public class Sysinfo {
    int cores = 1;
    int mhz = 1000;
    int memMeg = 256;
    String hostName = "unknown";

    /**
     * Constructor/init. Sets the variables.
     * Call getCores etc after this.
    */
    public Sysinfo() {
	String osName = System.getProperty("os.name");
        if (osName.startsWith("Windows")) {
            regReadCpu();
            wmicGetMem();
	}
        if (osName.startsWith("Linux")) {
            procReadCpu();
            procReadMem();
	}
        if (osName.startsWith("Mac")) {
	    sysctlReadCpuMem();
	}
        readHostName(); //this should work for all OS's
    }

    public int getCores() {
        return cores;
    }

    public int getMhz() {
        return mhz;
    }

    public int getMemMeg() {
        return memMeg;
    }

    public long getFreeDiskMeg() {
	long freeMegs = 0;
	//Java 1.5 does not support the getUsableSpace() function
        String cmd = "df -m .";
        String osName = System.getProperty("os.name");
	if (osName.startsWith("Windows")) {
            cmd = "cmd /c dir /-C";
        }
	String dfOut = org.natto.StaticTools.execGet(cmd, false);
        dfOut = dfOut.trim();
        String lastLine = "";
	int crPos = dfOut.lastIndexOf("\n");
        if (crPos > 0) {
	    lastLine = dfOut.substring(crPos + 1, dfOut.length());
        }
        String[] dfFigs = lastLine.split("\\s+");
        if (dfFigs.length > 3) {
            String freeStr = dfFigs[3];
            try {
                if (osName.startsWith("Windows")) {
                    //this is bytes
                    long freeBytes = Long.parseLong(freeStr);
                    long freeKs = freeBytes/1024;
                    freeMegs = freeKs/1024;
                } else {
                    freeMegs = Long.parseLong(freeStr);
                }
            } catch (Exception e) {}
	}
	return freeMegs;
    }

    public String getHostName() {
	return hostName;
    }

    /** should be called by constructor. Sets hostName.*/
    void readHostName() {
	try {
	    java.net.InetAddress localMachine
                 = java.net.InetAddress.getLocalHost();
	    hostName = localMachine.getHostName();
	}
	catch (Exception e) { e.printStackTrace(); }
    }

    /* get stuff from /proc/cpuinfo, Linux style. */
    void procReadCpu() {
        try {
            BufferedReader br = new BufferedReader(
                  new FileReader("/proc/cpuinfo"));
	    StringBuilder sb = new StringBuilder();
	    String line = br.readLine();
            int myCores = 0;
	    while (line != null) {
                if (line.startsWith("processor")) {
		    myCores++;
		}
                if (line.contains("cpu MHz")) {
		    int afterCo = line.indexOf(": ");
                    if (afterCo > 0) {
                        String mHzStr = line.substring(afterCo + 2,
						  line.length());
			float mFloat = Float.parseFloat(mHzStr);
                        mhz = (int)mFloat;
		    }
		}
		line = br.readLine();
	    }
	    br.close();
            cores = myCores;
	} catch (Exception e) { e.printStackTrace(); }
        //System.out.println(cores + " cores @ " + mhz);
    }

    /* get stuff from /proc/meminfo, Linux style. */
    void procReadMem() {
        try {
            BufferedReader br = new BufferedReader(
                  new FileReader("/proc/meminfo"));
	    StringBuilder sb = new StringBuilder();
	    String line = br.readLine();
	    while (line != null) {
                if (line.contains("MemTotal:")) {
		    //this line looks like:
                    //"MemTotal:    NUM     kb
                    String strParts[] = line.split("\\s+");
                    if (strParts.length == 3) {
			int mKb = Integer.parseInt(strParts[1]);
                        memMeg = mKb/1024;
		    }
		}
		line = br.readLine();
	    }
	    br.close();
	} catch (Exception e) { e.printStackTrace(); }

    }

    void regReadCpu() {
        String cmd = "reg query " +
                "HKLM\\HARDWARE\\DESCRIPTION\\System\\CentralProcessor /s";
        try {
            // Run reg query, then read output with
            // StreamReader (internal class)
            String output = org.natto.StaticTools.execGet(cmd, false);
            //parse the output to obtain core count and frequency
            String[] lines = output.split("\\n");
            int myCores = 0;
            for (int i=0; i<lines.length; i++) {
		String line = lines[i];
                if (line.contains("CentralProcessor\\")) {
		    myCores++;
		}
                if (line.contains("~MHz")) {
                    //split to get the 0x part that is 3rd
                    String[] lineParts = line.split("\\s+");
                    if (lineParts.length == 4) {
                        String zeroStr = lineParts[3];
                        if (zeroStr.startsWith("0x")) {
                            String noZero = zeroStr.substring(2,
					    zeroStr.length());
                            mhz = Integer.parseInt(noZero, 16);
			}
		    }
		}
	    }
            cores = myCores;
            System.out.println(cores + " cores @ " + mhz + " MHz");
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /** Use the wmic memorychip get capacity command.
     * Note: this does not work on XP home edition that does not have wmic,
     * but we just use the default.
     */
    void wmicGetMem() {
	long memTotBytes = 0;
        try {
	    String chipInfo =
                 org.natto.StaticTools.execGet("wmic memorychip get capacity",
                                               false);
	    String[] lines = chipInfo.split("\\n");
	    for (int i=0; i<lines.length; i++) {
		String line = lines[i].trim();
                if (line.length() > 0) {
		    if (Character.isDigit(line.charAt(0))) {
			//try to convert
			//System.out.println("Converting " + line);
			long lineMem = Long.parseLong(line);
			memTotBytes = memTotBytes + lineMem;
			long toMegs = 1024*1024;
			memMeg = (int)(memTotBytes / toMegs);
		    }
		}
	    }
	} catch (Exception e) { e.printStackTrace(); }
    }


   void sysctlReadCpuMem() {
       try {
	    String myInfo = org.natto.StaticTools.execGet("sysctl -n hw.ncpu",
                                                          false);
	    System.out.println(myInfo);
            if (myInfo != null) {
		//the result should be just one line, trim and convert
		myInfo = myInfo.trim();
		cores = Integer.parseInt(myInfo);
	    }
	} catch (Exception e) { e.printStackTrace(); }
       try {
	    String myInfo =
                    org.natto.StaticTools.execGet("sysctl -n hw.physmem",
                                                  false);
	    //System.out.println(myInfo);
            if (myInfo != null) {
		myInfo = myInfo.trim();
		long memBytes = Long.parseLong(myInfo);
		long toMegs = 1024*1024;
                long memLong = memBytes/toMegs;
                memMeg = (int) memLong;
	    }
	} catch (Exception e) { e.printStackTrace(); }
       try {
	    String myInfo =
                    org.natto.StaticTools.execGet("sysctl -n hw.cpufrequency",
                                                  false);
	    System.out.println("hw.cpufrequency " + myInfo);
            if (myInfo != null) {
		myInfo = myInfo.trim();
		long freqHz = Long.parseLong(myInfo);
                long toMhz = 1000 * 1000;
                long mhzLong = freqHz/toMhz;
                System.out.println(" now " + mhzLong);
                //in 10.6 Apple decided to have 000 000 extra
                if ((mhzLong > 1000000) && myInfo.endsWith("000000")) {
                    mhzLong = mhzLong/toMhz;
		}
                mhz = (int) mhzLong;
	    }
	} catch (Exception e) { e.printStackTrace(); }

   }

}
