/**
Copyright © 2014 Marko Niinimaki. All Rights Reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
package org.natto;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Calendar;
import java.util.Properties;
import java.util.TimeZone;

/**
 * Represents a task/job that exists in the Executor system.
 * Please note that tasks are not created by this class. You need to create
 * a job description, and the system will create a task using it.
 * @see JobDescription
 * @author man@cern.ch
 */
public class Task {

   /** when started */
   private long startTime;
   private long finishTime;
   private int taskStatus;
   private String runfileName;
   private String execHost;
   private Thread jobProcess;
   private int exitCode;
   public static int IDLE = 0;
   public static int RUNNING = 1;
   public static int DONE = 2;
   public static int NOTIFIED = 3;

   /**
    * Proper constructor to create a "run". See JobDescription to submit a task.
    * @param rfName run file name
    * @param myPr thread that runs the task
    * @param runsWhere info name of the host
    */
   public Task(String rfName, Thread myPr, String runsWhere) {
                   //check the time
       Calendar calendar = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
       long secSinceEpoch = calendar.getTimeInMillis() / 1000L;
       runfileName = rfName;
       jobProcess = myPr;
       startTime = secSinceEpoch;
       taskStatus = RUNNING;
       execHost = runsWhere;
       //write the .run file
       Properties runProperties = new Properties();
       runProperties.put("starttime", Long.toString(startTime));
       runProperties.put("status", Integer.toString(taskStatus));
       runProperties.put("exechost", execHost);
       try {
            runProperties.store(new FileOutputStream(rfName), null);
       } catch (Exception e) { e.printStackTrace(); }
   }

   /** Fake constructor, just set file name and return.
    * @param rfName run file name
    */
   public Task(String rfName) {
       runfileName = rfName;
   }

   /**
    * Remove all control files related to this task.
    */
   public void eraseCompletely() {
       String outFileName = runfileName.replace(".run", ".stdout");
       String errFileName = runfileName.replace(".run", ".stderr");
       String jdFileName  = runfileName.replace(".run", ".jd");
       String signedFileName = runfileName.replace(".run", ".jd.asc");
       String[] allFileNames = { runfileName, outFileName, errFileName,
           jdFileName, signedFileName
       };
       for (int i=0; i<allFileNames.length; i++) {
           File file = new File(allFileNames[i]);
           if (file.exists()) {
               file.delete();
           }
       }

   }

   public void finishTask(int exitCode) {
       Calendar calendar = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
       long secSinceEpoch = calendar.getTimeInMillis() / 1000L;
       finishTime = secSinceEpoch;
       Properties runProperties = new Properties();
       try {
           runProperties.load(new FileInputStream(runfileName));
           //add exit code
           runProperties.put("status", Integer.toString(DONE));
           runProperties.put("exitcode", Integer.toString(exitCode));
           runProperties.put("finishtime", Long.toString(finishTime));
           FileOutputStream fileOut = new FileOutputStream(runfileName);
           runProperties.store(new FileOutputStream(runfileName), null);
           fileOut.close();
       } catch (Exception e) { e.printStackTrace(); }
   }

   public int getStatus() {
       Properties runProperties = new Properties();
       int retMe = IDLE;
       try {
             runProperties.load(new FileInputStream(runfileName));
             String strStatus = runProperties.getProperty("status");
             retMe = Integer.parseInt(strStatus);
       } catch (Exception e) { }
       return retMe;
   }

   public void setStatus(int status) {
       taskStatus = status;
       Properties runProperties = new Properties();
       try {
             runProperties.load(new FileInputStream(runfileName));
             runProperties.put("status", Integer.toString(status));
             runProperties.store(new FileOutputStream(runfileName), null);
       } catch (Exception e) {}
    }

   public int getExitCode() {
       Properties runProperties = new Properties();
       int retMe = -100;
       try {
             runProperties.load(new FileInputStream(runfileName));
             String strStatus = runProperties.getProperty("exitcode");
             retMe = Integer.parseInt(strStatus);
             runProperties.store(new FileOutputStream(runfileName), null);
       } catch (Exception e) { }
       return retMe;

   }

}
