/**
Copyright © 2014 Marko Niinimaki. All Rights Reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
package org.natto;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.StringReader;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Properties;
import javax.swing.JOptionPane;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

/**
 * Represents executing/observing a task as a thread in the current computer.
 * @author man@cern.ch
 */
public class TaskExecution {

    /** Starts a command in a thread as a sub-shell of the current shell.
     * The run file, stdout and stderr of the execution are written in a file
     * with the same base name as the job description.
     * @param dir execute in this directory
     * @param cmd the command to exec the file
     * @param fileN the file to exec
     * @param args command line arguments
     * @param baseOutput base name for stdout/stderr files
     * @param interval time (ms) to wait between probing i/o
     * @param myId this host's id
     * @return the launched process thread
     */
    public static Thread execCommand(final String dir, final String cmd,
                                  final String fileN,
                                  final String args, final String baseOutput,
                                  final int interval,
                                  final String myId) {
        Thread processThread = new Thread(new Runnable() {

           private String getLineFromFileThenDelete(File fileToRead) {
               String retMe = "\n";
               try {
                   if (fileToRead.exists() && (fileToRead.length() > 0)) {
                       FileInputStream fstream =
                               new FileInputStream(fileToRead);
                       DataInputStream in = new DataInputStream(fstream);
                       BufferedReader br =
                               new BufferedReader(new InputStreamReader(in));
                       retMe = br.readLine();
                       fileToRead.delete();
                    }
               } catch (Exception e) {}
               return retMe;
            }

            public void run() {
                Process process = null;
                String osName = System.getProperty("os.name");
                String fullCmd = cmd + " " + fileN;
                if (cmd.startsWith("./")) {
                    //this is an uploaded executable, we do not need the file
                    //but we need to chmod it
                    String chRes = StaticTools.execGet("chmod +x "
                                     + dir + "/" +fullCmd, true);
                    fullCmd = cmd;
                }
                String cmArgs = "";
                if (args != null) { cmArgs = " " + args; }
                String[] command = {"bash", "-c", fullCmd + cmArgs};
                if (osName.startsWith("Windows")) {
                    String[] winc = {"CMD", "/Q", "/C", fullCmd + cmArgs};
                    command = winc;
                }
                System.out.println("execCommand " + fullCmd + cmArgs);
                if (fullCmd.startsWith("INTERNAL")) {
                    handleInternal(cmd, args, dir, baseOutput, myId);
                    //tell that we are done
                    String runFileName = dir + "/" + baseOutput + ".run";
                    Task finishedTask = new Task(runFileName);
                    finishedTask.finishTask(0);
                    return;
                }
                ProcessBuilder probuilder = new ProcessBuilder(command);
                //run file
                String toFileName = dir + "/" + baseOutput + ".to";
                File toFile = new File(toFileName); //this file contains input
                                                    //to the process
                //Set up  work directory
                probuilder.directory(new File(dir));
                try {
                   process = probuilder.start();
                   //Create a stream that goes into the process if the
                   //process needs to read stdin
                   OutputStream toProcess = process.getOutputStream();
                   DataOutputStream dos = new DataOutputStream(toProcess);

                   //Get outputs
                   InputStream is = process.getInputStream();
                   InputStream es = process.getErrorStream();
                   InputStreamReader isr = new InputStreamReader(is);
                   InputStreamReader esr = new InputStreamReader(es);
                   BufferedReader br = new BufferedReader(isr);
                   BufferedReader bre = new BufferedReader(esr);
                   String line;

                   //stdout of the job
                   BufferedWriter fout = new BufferedWriter(
                        new FileWriter(new File(dir + "/"
                                + baseOutput + ".stdout")));

                   while ((line = br.readLine()) != null) {
                       if (line.startsWith("ASK")) {
                           //the process may stop here! We need add a time
                           //and wait for the reply
                           long epoch = StaticTools.getEpoch();
                           line = line + " " + epoch;
                           fout.write(line + "\n");
                           fout.flush();
                           System.out.println("WAITING FOR ANSWER " + line);
                           boolean noAnswer = true;
                           while (noAnswer) {
                                Thread.sleep(interval);
                                String reply = getLineFromFileThenDelete(toFile);
                                if (reply.equals("")|| reply.equals("\n")) {
                                } else {
                                    //Got answer, push
                                    dos.writeUTF(reply+"\n");
                                    dos.flush();
                                    noAnswer = false;
                                }
                           }
                       } else {
                           fout.write(line + "\n");
                           fout.flush();
                       }
                   }
                   fout.close();

                   //stderr of the job
                   BufferedWriter ferr = new BufferedWriter(
                        new FileWriter(new File(dir + "/"
                                + baseOutput + ".stderr")));
                   while ((line = bre.readLine()) != null) {
                        ferr.write(line + "\n");
                   }
                   ferr.close();
               } catch (Exception e) {
                       e.printStackTrace();
               }
               //wait until process finished, then write exit code
               boolean goOn = true;
               int retVal = Integer.MIN_VALUE;
               while (goOn) {
                   if (process == null) {
                       goOn = false;
                       retVal = -100;
                   }
                   try {
                       retVal = process.exitValue();
                   } catch (Exception e) { }
                   //System.out.println("rv " + retVal);
                   if (retVal > Integer.MIN_VALUE) {
                       goOn = false;
                       String runFileName = dir + "/" + baseOutput + ".run";
                       Task finishedTask = new Task(runFileName);
                       finishedTask.finishTask(retVal);
                   } else {
                       try {
                           Thread.sleep(interval);
                       } catch (Exception e) { }
                   }
               }
           } //run



        }); //Thread
        processThread.start();
        return processThread;
   }

   /**
   * Execute an internal command like XMLLS in this host.
   * @param cmd command to execute
   * @param args command arguments
   * @param dir work directory of the executor
   * @param baseOutput job file name without extension
   * @param myId this host's id
   */
   private static void handleInternal(String cmd, String args, String dir,
                                       String baseOutput, String myId) {
       File outFile = new File(dir + "/" + baseOutput + ".stdout");
       if (cmd.equals("INTERNAL:XMLLS")) {
           //get the param
           if ((args == null) || ("$HOME".equals(args))) {
                args = System.getProperty("user.home");
           }
           //get the list of files there
           File argDir = new File(args);
           String fileNames[] = argDir.list();
           //write into stdout

           try {
               FileWriter fwr = new FileWriter(outFile);
               fwr.write("<executor host=\""+ myId +"\">\n<dirlist directory=\""
                                      + argDir + "\">\n");
               Arrays.sort(fileNames);
               for (int i=0; i<fileNames.length; i++) {
               if (!(fileNames[i].startsWith("."))) {
                   File tmpFile = new File(argDir + "/" + fileNames[i]);
                   String fileType = "file";
                   if (tmpFile.isDirectory()) {
                       fileType = "directory";
                   }
                   fwr.write("<entry type=\"" + fileType
                                  + "\">" + fileNames[i] + "</entry>\n");
                   }
               }
               fwr.write("</dirlist>\n</executor>\n");
               fwr.close();
           } catch (Exception e) {}
       }
       if (cmd.equals("INTERNAL:FILECOPY")) {
           try {
               String errMsg = "";
               //FILECOPY takes 2 arguments, first is in quotes
               int latterQpos = args.indexOf("\" ");
               if (latterQpos > 2) {
                   //remove quotes
                   String copyThis = args.substring(1, latterQpos).trim();
                   String toThis = args.substring(latterQpos + 1).trim();
                   if (".".equals(toThis)) { toThis = dir; }
                   System.out.println("Copying " + copyThis + " to " + toThis);
                   File fileForCopy = new File(copyThis);
                   boolean copyOK = StaticTools.copyFileIfNeeded(fileForCopy,
                                                                 toThis);
                   if (!(copyOK)) { errMsg = "Copy failed."; }
               } else {
                   errMsg = "Arguments are invalid: " + args;
               }
               //notify by stdout
               if ("".equals(errMsg)) { errMsg = "File copy OK" ; }
               FileWriter fwr = new FileWriter(outFile);
               fwr.write(errMsg + "\n");
               fwr.close();
           } catch (Exception e) {}
       }
   }

   /**
   * Keep track of lines in output files.
   */
   static HashMap jobAndLine = new HashMap();

   /** Show messages and handle the finish of jobs that I have submitted.
     * If job has finished, notify and change name.
     * @param myDir directory
     * @param fileName the file that finishes with .run
     * @param myGui GUI to use for notification
     * @param secOn is security on?
    */
    public static void notifyAboutSubmittedJobs(String myId, int interval,
                                     String myDir, String fileName,
                                     GenGui myGui, boolean secOn) {
        int lastAskLine = 0; //what was the line of the last question in output
	//check the name without extension
        String noExt = fileName.replace(".run", "");
	Task ohJob = new Task(myDir + "/" + fileName);
        int ohJobStat = ohJob.getStatus();
        if (ohJobStat == Task.NOTIFIED) {
            return;
        }

        String jdFileName = myDir + "/" + noExt + ".jd";
        if (secOn) { jdFileName = jdFileName + ".asc"; }
        //check in the jd file if it was submitted by me
        Properties jobProp =
                org.natto.StaticTools.getPropertiesFromFile(jdFileName);
        String submitHost = (String)jobProp.get("clientid");
        //System.out.println("I am " + myId + " submitter is " + submitHost);
        if (!(myId.equalsIgnoreCase(submitHost))) { return; }

        String fullOutFn = myDir + "/" + noExt + ".stdout";
        String fullErrFn = myDir + "/" + noExt + ".stderr";
        String fullDiscFn = myDir + "/" + noExt + ".to"; //to process
        File outFile = new File(fullOutFn);
        int exitInFile = 0;
	if (ohJobStat == Task.DONE) {
	    try {
                //wait 1 more interval so that the process has time to
                //write the last TELL if it needs to
                exitInFile = ohJob.getExitCode();
                Thread.sleep(interval);
           } catch (Exception e) {}
        }
        try {
           //check lines that we need to show to user
           if (outFile.exists()) {
               //get the last line from file
               FileInputStream fstream =
                               new FileInputStream(outFile);
               String lastLine = "";
               String curLine;
               DataInputStream in = new DataInputStream(fstream);
               BufferedReader br =
                       new BufferedReader(new InputStreamReader(in));
               int lineNo = 0;
               while ((curLine = br.readLine()) != null) {
                   if (curLine.startsWith("TELL ") ||
                            curLine.startsWith("ASK ")) {
                      lastLine = curLine;
                      lineNo++;
                   }
               }
               in.close();
               if (lastLine.startsWith("TELL ")) {
                   myGui.doTell(lastLine);
               }
               if (lastLine.startsWith("ASK ")) {
                   //did we ask this already?
                   if (jobAndLine.containsKey(fullOutFn)) {
                       lastAskLine = (Integer)jobAndLine.get(fullOutFn);
                    }
                    if (lineNo != lastAskLine) {
                        //remove the timestamp at the end
                        lastLine = lastLine.substring(0,
                                lastLine.lastIndexOf(" "));
                        String reply = myGui.doAsk(lastLine);
                        //write the reply to a discussion file that
                        //goes to the process
                        File toProcFile = new File(fullDiscFn);
                        FileWriter toProcFw = new FileWriter(toProcFile);
                        toProcFw.write(reply + "\n");
                        toProcFw.close();
                        jobAndLine.put(fullOutFn, lineNo);
                   }
               }

            }
        } catch (Exception e) {}

        //finally: notify user, handle output from internal jobs
	if (ohJobStat == Task.DONE) {
            //get the stdout and stderr files
            String myOut = StaticTools.getFileContent(fullOutFn, 1);
            String myErr = StaticTools.getFileContent(fullErrFn, 1);
	    try {
                ohJob.setStatus(Task.NOTIFIED);
	        System.out.println("My job has finished with code "
                                + exitInFile);
                String myNotif = "Stdout starts with " + myOut;
                if (exitInFile > 0) {
                    myNotif = "Stderr starts with " + myErr;
                }
                if (myOut.startsWith("<executor ")) {
                    //show internal res
                    myOut = StaticTools.getFileContent(fullOutFn);
                    RenderedDirlist myRdl = XMLRenderDirlist(myOut);
                    String[] toShow = myRdl.getEntries();
                    boolean[] myDirs = myRdl.getDirData();
                    String dirListed = myRdl.getDirName();
                    String execHost = myRdl.getHostName();
                    org.natto.swingGUI.DirPresenter dirPres =
                            new org.natto.swingGUI.DirPresenter(toShow, myDirs);

                    int selInd = dirPres.getselectedIndex();
                    if (selInd > -1) {
                        String selItem = toShow[selInd];
                        String nextItem = dirListed + "/" + selItem;
                        String osName = System.getProperty("os.name");
                        if (osName.startsWith("Windows")) {
                            nextItem = dirListed + "\\" + selItem;
                            nextItem = nextItem.replace("\\", "\\\\");
                        }
                        JobDescription newJd = new JobDescription(myDir);
                        if (myDirs[selInd]) {
                            //generate command for listing this directory
                            newJd.writeInternalCommand("XMLLS", nextItem,
                                                        execHost);
                        } else {
                            //it is a file
                            int reply = JOptionPane.showConfirmDialog(null,
                               "Do you want to copy file\n" + selItem
                               + " to directory\n"
                               + myDir + "?", null,
                               JOptionPane.YES_NO_OPTION);
                            if (reply == JOptionPane.YES_OPTION) {
                                 //command for file copy
                                 newJd.writeInternalCommand("FILECOPY",
                                         "\"" + nextItem + "\" .", execHost);
                            }
                        }
                    }

                    ohJob.eraseCompletely();
                } else {
                    myGui.showJobFinishDiag(myNotif, myDir, exitInFile);
                }
            } catch (Exception e) {  e.printStackTrace(); }
	}
    }

    /**
     * Generates a string that can be displayed in a panel.
     * @param xmlSource source, first tag <executor>
     * @return string that can be shown
     */
    static RenderedDirlist XMLRenderDirlist(String xmlSource) {
        if (xmlSource == null) { return null; }
        if (!(xmlSource.contains("<dirlist"))) { return null;  }
        //actual parsing
        boolean[] isDir = null;
        String[] fileNames = null;
        String execHost = "";
        String dirName = "";
        DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
        try {
            DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
            InputSource is = new InputSource(new StringReader(xmlSource));
            Document doc = dBuilder.parse(is);
            //get host name
            int hostPos = xmlSource.indexOf(" host=");
            if (hostPos > 0) {
                int endPos = xmlSource.indexOf("\"", hostPos + 8);
                execHost = xmlSource.substring(hostPos + 7, endPos);
            }
            //get directory name
            int dirPos = xmlSource.indexOf(" directory=");
            if (dirPos > 0) {
                int endPos = xmlSource.indexOf("\"", dirPos + 13);
                dirName = xmlSource.substring(dirPos + 12, endPos);
            }
            //get entries
            NodeList nList = doc.getElementsByTagName("entry");
            //contruct arrays
            isDir = new boolean[nList.getLength()];
            fileNames = new String[nList.getLength()];
            for (int i=0; i<nList.getLength(); i++) {
                String eName = nList.item(i).getTextContent();
                String eType = ((Element)nList.item(i)).getAttribute("type");
                boolean thisIsDir = eType.equals("directory");
                isDir[i] = thisIsDir;
                fileNames[i] = eName;
            }
        } catch (Exception e) { e.printStackTrace(); }
        RenderedDirlist rdl = new RenderedDirlist(fileNames, isDir,
                                                  dirName, execHost);
        return rdl;
    }

    static class RenderedDirlist {
        private String[] dirEntries;
        private boolean[] isDir;
        private String execHost;
        private String dirName;

        public RenderedDirlist(String[] names, boolean[] theseAreDirs,
                String dirN, String hostN) {
            dirEntries = names; isDir = theseAreDirs; dirName = dirN;
            execHost = hostN;
        }

        String[] getEntries() { return dirEntries; }
        boolean[] getDirData() { return isDir; }
        String getDirName() { return dirName; }
        String getHostName() { return execHost; }
    }
}
