/**
Copyright © 2014 Marko Niinimaki. All Rights Reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**/
package org.natto.swingGUI;
import java.awt.*;
import java.awt.event.*;
import java.util.HashMap;
import java.util.Set;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SwingConstants;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;

/**
 * Gui elements for the simple execution service.
 * @author man@cern.ch
 */
public class MainGui implements org.natto.GenGui {

    /** security manager needed to respond to the security button. **/
    static org.natto.ExecSecurity mySecurity =
            org.natto.ExecSecurity.getInstance();
    private final String secIsOffText = "<html>&nbsp;Security is off, tasks will be "
                                  + "executed"
                                  + "<br/>&nbsp;without checking their "
                                  + "signatures.</html>";
    private final String secIsOnText = "<html>&nbsp;Security is on, only valid, signed "
                                  + "tasks<br/>&nbsp;will be executed.</html>";
    private final String turnSecOnButtonText = "Turn security on";
    private final String turnSecOffButtonText = "Turn security off";

    /** The text of this will be in the main window.
     */
    private JLabel messageLabel = new JLabel();

    /** This is the default content of the messageLabel.
     * It is set if the message is "".
     */
    private String defaultMsg = "";
    private JLabel tellLabel = new JLabel(" "); //for update messages
    private JLabel secLabel = new JLabel(" ");
    private JButton secButton = null; //security on/off
    private JButton submitButton = null; //submitting
    private JButton spyButton = null;
    private JButton cButton = null; //closing
    private String mainDir = ""; //app's dir
    private HashMap infos = null; //for the submitgui

    /**
     * Creates a GUI window that keeps the application alive.
     * NB: the buttons will be inactive until you run
     * updateInfosActivate
     * @param appName name of the application.
     * @param myDir directory name where the application runs.
    */
    public MainGui(String appName, String myDir) {
        //thing to display if we cannot create GUI or tray.
        String myMsg = "Program " + appName + " is now running in directory"
                    + "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;"
                    + "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\n"
                    + myDir
                    + "\nuntil you close this application.";
        defaultMsg = myMsg;
        mainDir = myDir;
        try {
            new nonBlockingDialog("Executor", myMsg);
        } catch (Exception e) {
            e.printStackTrace();
            //very probably "no X". Just print the message.
            System.out.println(myMsg);
        }
    }

    /** update the info file hash map and set buttons active.
     * @param hm use this for host infos
    */
    public void updateInfosActivate(HashMap hm) {
        infos = hm;
        //if this is run without X, there are no buttons
        if (cButton != null) { cButton.setEnabled(true); }
        if (submitButton != null) { submitButton.setEnabled(true); }
        if (spyButton != null) { spyButton.setEnabled(true); }
        if (secButton != null) { secButton.setEnabled(true); }
    }

    /** Sets the text in the main window.
     * @param msg message
     */
    public void setMessage(String msg) {
        if ("".equals(msg)) { msg = defaultMsg; }
        String htmlMsg = msg.replace("\n", "<br/>");
        htmlMsg = "<html>" + htmlMsg + "</html>";
        try {
            messageLabel.setHorizontalAlignment(SwingConstants.LEFT);
            messageLabel.setText(htmlMsg);
        } catch (Exception e) {
            //probably environment where there is no X
            System.out.println(msg);
        }
    }

    /**
     * Sets the security on/off message and button text.
     * Based on the status of security manager.
     */
    public void setSecurityMessages() {
        try {
            if (mySecurity.securityOn()) {
                secLabel.setText(secIsOnText);
                secButton.setText(turnSecOffButtonText);
            } else {
                secLabel.setText(secIsOffText);
                secButton.setText(turnSecOnButtonText);
            }
        } catch (Exception e) { }

    }

    /** Shows files in the directory using a browser.
     * @param notif notification text
     * @param dirName directory to show
     * @param exitCode exit code
     */
    public void showJobFinishDiag(String notif, String dirName, int exitCode) {
        int optionType = JOptionPane.OK_CANCEL_OPTION;
        int res = JOptionPane.showConfirmDialog(null,
                "Task finished, code: " + exitCode + ".\n" + notif + "\n"
                + "Press OK to see the files",
                "Your task", optionType);
        if (res == JOptionPane.OK_OPTION) {
            try {
                org.natto.swingGUI.BareBonesBrowserLaunch.openURL(dirName);
            } catch (Exception e) { e.printStackTrace(); }
        }
    }

    /** Internal class to show a dialog that contains instructions,
     * remains floating in the screen and kills the application on close.
     */
    public class nonBlockingDialog extends JDialog implements ActionListener {
        /**
        * Constructor.
        * @param title window title
        * @param message message. We support multi-line.
        */
        public nonBlockingDialog(String title, String message) {
            super(new JFrame(), title, true);
            JMenuBar menubar = new JMenuBar();
            JMenu helpMenu = new JMenu("Help and examples");
            menubar.add(helpMenu);
            setJMenuBar(menubar);
            myMenuListener mml = new  myMenuListener();
            helpMenu.addMenuListener(mml);
            JPanel messagePane = new JPanel();
            getContentPane().add(messagePane, BorderLayout.NORTH);
            messagePane.add(messageLabel);
            setMessage(message);
            //discussion panel in the middle
            JPanel discPane = new JPanel(new BorderLayout());
            discPane.add(secLabel, BorderLayout.WEST);
            secButton = new JButton("");
            secButton.addActionListener(this);
            setSecurityMessages();
            discPane.add(secButton, BorderLayout.EAST);
            discPane.add(tellLabel, BorderLayout.SOUTH);
            getContentPane().add(discPane, BorderLayout.CENTER);
            //buttons at the bottom
            JPanel buttonPane = new JPanel();
            cButton = new JButton("Close");
            spyButton = new JButton("See hosts");
            JLabel empty = new JLabel("<html>&nbsp;&nbsp;&nbsp;&nbsp;"
                                      + "&nbsp;&nbsp;&nbsp;&nbsp;</html>");
            submitButton = new JButton("Submit");
            this.setModal(false);
            buttonPane.add(submitButton);
            buttonPane.add(spyButton); //later
            buttonPane.add(empty);
            buttonPane.add(cButton);
            submitButton.addActionListener(this);
            spyButton.addActionListener(this);
            cButton.addActionListener(this);
            //to start, the buttons are inactive
            submitButton.setEnabled(false);
            cButton.setEnabled(false);
            secButton.setEnabled(false);
            spyButton.setEnabled(false);
            getContentPane().add(buttonPane, BorderLayout.SOUTH);
            setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
            pack();
            setVisible(true);
        }

        /** Control actions of all the buttons.
         * @param actEv event launched by buttons.
         */
        public void actionPerformed(ActionEvent actEv) {
            String alertMsg = "";
            if (actEv.getSource() == cButton) {
                setVisible(false);
                dispose();
                System.exit(0);
            }
            if (actEv.getSource() == secButton) {
                //check if we need to turn on or off
                if (mySecurity.securityOn()) {
                    mySecurity.turnOff();
                } else {
                    alertMsg = mySecurity.turnOn();
		    String sigNers = mySecurity.getSigners();
                    if (!(sigNers.equals(""))) {
                        JOptionPane.showMessageDialog(null,
                                                      "Accepting tasks from:\n"
						      + sigNers);
		    }
                }
                if (!("".equals(alertMsg))) {
                    JOptionPane.showMessageDialog(null, alertMsg);
                }
                JOptionPane.showMessageDialog(null, "Please make sure you\n"
                        + "understand how this works by checking\n"
                        + "Help and Examples in the menu.");
                setSecurityMessages();
            }
            if (actEv.getSource() == submitButton) {
                //bring up submit dialog
                //JOptionPane.showMessageDialog(null, "TBD");
                org.natto.swingGUI.SubmitGui subGui =
                   new org.natto.swingGUI.SubmitGui(mainDir, infos, "");
            }
            if (actEv.getSource() == spyButton) {
                //System.out.println("spy");
                //Get the host names from infos
                Set keys = infos.keySet();
                Object[] keyObjVals = keys.toArray();
                String[] keyVals = new String[keyObjVals.length];
                for (int i = 0; i < keyObjVals.length; i++) {
                    //these are keys. Get the corresponding name value.
                    String hostInfoKey = keyObjVals[i].toString();
                    String n = org.natto.Infofiles.getHostProperty(infos,
                                                        hostInfoKey,
                                                        "name");
                    keyVals[i] = n + " (" + hostInfoKey + ")";
                }
                //bring up host select dialog
                org.natto.swingGUI.hostSelectDialog hSel =
                   new org.natto.swingGUI.hostSelectDialog(null, true, keyVals);
                String[] hostAction = hSel.getSelectedHostAndAction();
                String toHost = hostAction[0]; String toAction = hostAction[1];
                if (toHost != null) {
                    if (toAction.equals("see")) {
                        //insert code to run an "ls" job here
                        org.natto.jobDescription jDesc =
                                new org.natto.jobDescription(mainDir);
                        jDesc.writeInternalCommand("XMLLS", "$HOME", toHost);
                    } else {
                        org.natto.swingGUI.SubmitGui subGui =
                        new org.natto.swingGUI.SubmitGui(mainDir, infos, toHost);
                    }
                }
            }
        }
   } //nonblocking

    /** Display a question, request reply.
     * Note: formatting to cut too long lines.
     * @param myQuestion the question
     * @return the user's reply
     */
    public String doAsk(String myQuestion) {
        int okLen = 40;
        String[] parts = myQuestion.split(" ");
        String tmp = "";
        int thisLineLen = 0;
        for (int i = 0; i < parts.length; i++) {
            int thisPartLen = parts[i].length();
            tmp = tmp + " " + parts[i];
            thisLineLen = thisLineLen + thisPartLen;
            if (thisLineLen > okLen) {
                tmp = tmp + "\n";
                thisLineLen = 0;
            }
        }
        String s = (String) JOptionPane.showInputDialog(null, tmp);
        return s;
    }

    /** Show a message e.g progress.
     * @param myNote thing to show
     **/
    public void doTell(String myNote) {
        //System.out.println(myNote);
        tellLabel.setText(" " + myNote);
    }

      class myMenuListener implements MenuListener {
        public void menuCanceled(MenuEvent e) { }
        public void menuSelected(MenuEvent e) {
            //System.out.println("sel");
            try {
                String toPage =
                   "https://dl.dropboxusercontent.com/u/919611/executor.html";
                org.natto.swingGUI.BareBonesBrowserLaunch.openURL(toPage);
            } catch (Exception ex) { ex.printStackTrace(); }
        }
        public void menuDeselected(MenuEvent e) { }
   }

}